from typing import Any, Dict

from objectmodel.base import FieldABC, ObjectModelABC


__all__ = [
    'ObjectModel',
    'ObjectModelMeta'
]


def is_instance_or_subclass(val, klass) -> bool:
    try:
        return issubclass(val, klass)
    except TypeError:
        return isinstance(val, klass)


def _iter_fields(attrs: Dict[str, Any], field_class: type = FieldABC):
    for attr_name, attr in attrs.items():
        if is_instance_or_subclass(attr, field_class):
            yield attr_name, attr


class ObjectModelMeta(type):
    FIELDS_ATTR = '__fields__'

    def __new__(mcs, name, bases, attrs):

        cls_fields = dict(_iter_fields(attrs))

        # TODO: clear fields from actual instance? and override __getattr__?
        """for field_name in cls_fields:
            del attrs[field_name]"""

        cls = super().__new__(mcs, name, bases, attrs)

        # TODO: Use MRO instead of iteration over base classes?
        for base in bases:
            base_attrs = getattr(base, '__fields__', base.__dict__)
            for field_name, field in _iter_fields(base_attrs):
                cls_fields[field_name] = field

        cls.__fields__ = cls_fields
        return cls


class ObjectModel(ObjectModelABC, metaclass=ObjectModelMeta):
    DICT_FACTORY = dict

    __slots__ = '__state__'

    # fields class attr is set during class construction in ObjectModelMeta.__new__
    __fields__: Dict[str, FieldABC]

    def __init__(self, **kwargs):
        self.__state__ = {}
        for attr_name, attr_value in kwargs.items():
            if attr_name not in self.__fields__:
                raise AttributeError(f'Unexpected argument: {attr_name}, '
                                     f'no such field in model {self.__class__.__name__}')
            setattr(self, attr_name, attr_value)

        # TODO: Fix double validation (first one happens in the settattr)
        self.validate()

    def validate(self):
        for field in self.__fields__.values():
            if field.has_value(self) or field.required:
                value = field.__get__(self, self.__class__)
                field.validate(self, value)

    def deserialize(self, data: Dict[str, Any]):
        for key, value in data.items():
            try:
                field = self.__fields__[key]
                field.deserialize(self, value)
            except KeyError:
                # We've received some additional info that does not correspond to a field
                pass

    def serialize(self) -> Dict[str, Any]:
        return self.DICT_FACTORY(
            (field.name, field.serialize(self))
            for field in self.__fields__.values()
            if field.can_provide_value(self)
        )

    def clear(self):
        for field in self.__fields__.values():
            field.clear(self)

    def __setstate__(self, state: Dict[str, Any]):
        self.deserialize(state)

    def __getstate__(self) -> Dict[str, Any]:
        return self.serialize()

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(
                f'{name}={val!r}'
                for name, val in self.__state__.items()
            )
        )



from __future__ import annotations

import sys

from typing import Any, TypeVar, Type, Union, Optional, Callable

from objectmodel.base import ObjectModelABC, FieldABC
from objectmodel.errors import FieldValidationError, FieldValueRequiredError


__all__ = [
    'NOT_PROVIDED',
    'Field',
    'ObjectField',
    'ListCollectionField',
    'DictCollectionField',
    'ProxyField']


class _NotProvided:
    def __bool__(self):
        return False

    def __copy__(self):
        return self

    def __deepcopy__(self, _):
        return self

    def __repr__(self):
        return "<objectmodel.NOT_PROVIDED>"


NOT_PROVIDED = _NotProvided()

T = TypeVar('T')


class Field(FieldABC):
    __slots__ = 'name', 'default', 'required', 'allow_none', 'validator'

    def __init__(self,
                 name: str = NOT_PROVIDED,
                 default: Union[Callable[[], T], T] = NOT_PROVIDED,
                 required: bool = False,
                 allow_none: bool = False,
                 validator: Optional[
                     Callable[[Optional[ObjectModelABC], FieldABC, Any], None]
                 ] = None):
        self.name = name
        self.default = default
        self.required = required
        self.allow_none = allow_none
        self.validator = validator

        # Defaults also should be validated!
        if not callable(default):
            self.validate(None, default)

    def __get__(self, instance: ObjectModelABC, owner: Type[ObjectModelABC]) -> T:
        assert isinstance(instance, ObjectModelABC)
        try:
            return instance.__state__[self.name]
        except KeyError:
            if self.default is not NOT_PROVIDED:
                default = self.default
                if callable(default):
                    default = default()
                self.__set__(instance, default)
                return default
        raise FieldValueRequiredError(instance, self)

    def __set__(self, instance: ObjectModelABC, value: T):
        assert isinstance(instance, ObjectModelABC)
        self.validate(instance, value)
        instance.__state__[self.name] = value

    def __set_name__(self, owner, name):
        if self.name is NOT_PROVIDED:
            assert name and isinstance(name, str), 'String name must be specified'
            self.name = name

    def __delete__(self, instance):
        assert isinstance(instance, ObjectModelABC)
        if self.required:
            raise FieldValueRequiredError(instance, self)
        del instance.__state__[self.name]

    def serialize(self, instance: ObjectModelABC) -> Any:
        return self.__get__(instance, instance.__class__)

    def deserialize(self, instance: ObjectModelABC, value):
        self.__set__(instance, value)

    def has_default(self) -> bool:
        return self.default is not NOT_PROVIDED

    def has_value(self, instance: ObjectModelABC):
        return self.name in instance.__state__

    def can_provide_value(self, instance: ObjectModelABC):
        return self.default is not NOT_PROVIDED or self.name in instance.__state__

    def validate(self, model_instance: Optional[ObjectModelABC], value: T):
        if value is None and not self.allow_none:
            raise FieldValidationError(model_instance, self, value,
                                       'Cannot be None (allow_none=False)')
        if self.validator:
            value = self.__get__(model_instance, model_instance.__class__)
            self.validator(model_instance, self, value)

    def clear(self, instance):
        self.__delete__(instance)

    def __repr__(self):
        return '{}(name={!r}, default={!r}, required={!r}, allow_none={!r}, validator={!r})'\
            .format(
                self.__class__.__name__,
                self.name,
                self.default,
                self.required,
                self.allow_none,
                self.validator
            )


class ProxyField(Field):
    def __init__(self, attr_name: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._attr_name = attr_name

    def __get__(self, instance, owner):
        return getattr(instance, self._attr_name)

    def has_value(self, instance: ObjectModelABC) -> bool:
        return True


class ObjectField(Field):
    __slots__ = '_model'

    def __init__(self, name: str, model: type, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        assert issubclass(model, ObjectModelABC)
        self._model = model

    def serialize(self, instance: ObjectModelABC) -> Any:
        value = super().serialize(instance)
        if value is not None:
            return value.serialize()
        return None

    def deserialize(self, instance: ObjectModelABC, value):
        if value is not None:
            obj = self._model()
            obj.deserialize(value)
            super().deserialize(instance, obj)

    def validate(self, model_instance: ObjectModelABC, value):
        super().validate(model_instance, value)
        if not self.allow_none and value is not None:
            if not isinstance(value, ObjectModelABC):
                raise FieldValidationError(model_instance, self, value,
                                           f'Value should be of type: \'ObjectModel\'')
            value.validate()


class ListCollectionField(Field):
    __slots__ = '_model'

    def __init__(self, item_model: Union[str, Type[ObjectModelABC]], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._model = item_model

    def serialize(self, instance: ObjectModelABC) -> Any:
        value = super().serialize(instance)
        return [v.serialize() for v in value]

    def deserialize(self, instance: ObjectModelABC, value):
        deserialized_list = []
        item_cls = self._resolve_item_type()
        for v in value:
            obj = item_cls()
            obj.deserialize(v)
            deserialized_list.append(obj)
        super().deserialize(instance, deserialized_list)

    def _resolve_item_type(self) -> Type[ObjectModelABC]:
        if issubclass(self._model, ObjectModelABC):
            return self._model
        elif isinstance(self._model, str):
            self._model = getattr(sys.modules[__name__], self._model)
            return self._model
        raise TypeError(f'Cant resolve item model type: {self._model}')

    def validate(self, model_instance: ObjectModelABC, value):
        super().validate(model_instance, value)
        if not self.allow_none and value is not None:
            if not isinstance(value, list):
                raise FieldValidationError(model_instance, self, value,
                                           'Value should be of type: List[ObjectModel]')
            for item in value:
                if not isinstance(item, ObjectModelABC):
                    raise FieldValidationError(model_instance, self, value,
                                               f'List item {item!r} '
                                               f'should be of type: \'ObjectModel\'')
                item.validate()


class DictCollectionField(Field):
    __slots__ = '_model', '_dict_factory'

    def __init__(self, name: str, item_model: type, dict_factory: callable = dict,
                 *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        assert issubclass(item_model, ObjectModelABC)
        self._model = item_model
        self._dict_factory = dict_factory

    def serialize(self, instance: ObjectModelABC) -> Any:
        value = super().serialize(instance)
        return {k: v.serialize() for k, v in value.items()}

    def deserialize(self, instance: ObjectModelABC, value):
        deserialized_dict = self._dict_factory()
        for k, v in value.items():
            obj = self._model()
            obj.deserialize(v)
            deserialized_dict[k] = obj
        super().deserialize(instance, deserialized_dict)

    def validate(self, model_instance: ObjectModelABC, value: Any):
        super().validate(model_instance, value)
        if not isinstance(value, dict):
            raise FieldValidationError(model_instance, self, value,
                                       'Value should be of type Dict[ObjectModel]')
        for item in value.values():
            item.validate()



from typing import Any, Callable, TYPE_CHECKING

from objectmodel.errors import FieldValidationError

if TYPE_CHECKING:
    from objectmodel.base import ObjectModelABC, FieldABC


Validator = Callable[[ObjectModelABC, FieldABC, Any], None]


class FieldValidator:
    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        raise NotImplementedError()


class OfType(FieldValidator):
    def __init__(self, typ, allow_none=False):
        self.typ = typ
        self.can_be_null = allow_none

    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        if value is None:
            return self.can_be_null
        return isinstance(value, self.typ)


class NotEmptyString(FieldValidator):
    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        if not isinstance(value, str) or not value:
            raise FieldValidationError(instance, field, value,
                                       'Value should be a not-empty string')


class Numeric(FieldValidator):
    def __init__(self, allow_float=True, allow_none=False):
        self.allow_float = allow_float
        self.allow_none = allow_none

    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        if value is None:
            return self.allow_none
        if isinstance(value, float) and not self.allow_float:
            return False
        return isinstance(value, (float, int))


class PositiveNumeric(Numeric):
    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        super()(instance, field, value)
        return value >= 0


class MoreThanOrEqual(FieldValidator):
    def __init__(self, n):
        self.n = n

    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        return value >= self.n


class ItemsOfType(FieldValidator):
    def __init__(self, typ):
        self.typ = typ

    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        for item in value:
            if not isinstance(item, self.typ):
                return False
        return True


class ValidItems(FieldValidator):
    def __init__(self, *item_validators: Validator):
        self.item_validators = item_validators

    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        for i, item in enumerate(value):
            for validator in self.item_validators:
                validator(instance, field, item)
        return True


class NotEmptyList(FieldValidator):
    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        if not isinstance(value, list):
            return False
        return len(value) > 0


class MaxLen(FieldValidator):
    def __init__(self, max_len: int):
        self.max_len = max_len

    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        return len(value) <= self.max_len


class ChainValidator(FieldValidator):
    def __init__(self, *validators: FieldValidator):
        self.validators = validators

    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        for v in self.validators:
            v(instance, field, value)


class ValueIn(FieldValidator):
    def __init__(self, *args):
        if len(args) == 1:
            args = args[0]
        self.options = set(args)

    def __call__(self, instance: ObjectModelABC, field: FieldABC, value: Any):
        return value in self.options


is_not_empty_string = NotEmptyString()
is_not_empty_string_of_max_len = (lambda x: ChainValidator(is_not_empty_string, MaxLen(x)))
is_numeric = Numeric()
is_integer = Numeric(allow_float=False)
is_positive_numeric = PositiveNumeric()
is_positive_integer = PositiveNumeric(allow_float=False)
is_bool = OfType(bool)
is_of_type = OfType
is_list_of = (lambda typ: ChainValidator(OfType(list), ItemsOfType(typ)))
is_not_empty_list = NotEmptyList()
is_not_empty_list_of = (lambda typ: ChainValidator(is_not_empty_list, ItemsOfType(typ)))
value_in = ValueIn



from typing import Optional, Any

from objectmodel.base import ObjectModelABC, FieldABC


__all__ = [
    'FieldValidationError',
    'FieldValueRequiredError',
    'DuplicateFieldDefinitionError'
]


class FieldValidationError(AttributeError):
    """ Field validation error """

    def __init__(self, instance: Optional[ObjectModelABC], field: FieldABC, value: Any, message: str):
        super().__init__(f'Invalid value {value} for field {field!r} of {instance!r}: {message}')


class FieldValueRequiredError(AttributeError):
    """ Field is required but not set """

    def __init__(self, instance: ObjectModelABC, field: FieldABC):
        super().__init__(f'Field {field!r} of {instance!r} is not set')


class DuplicateFieldDefinitionError(AttributeError):
    """ A field with this name is already present in model """
    def __init__(self, field_name: str, class_name: str):
        super().__init__(f'Duplicate field definition found during {class_name} initialization, '
                         f'field: {field_name}')



from typing import Any, Type, Optional, Dict, TypeVar


__all__ = ['FieldABC', 'ObjectModelABC']


T = TypeVar('T')


class FieldABC:
    name: str
    required: bool

    def __get__(self, instance: 'ObjectModelABC', owner: Type['ObjectModelABC']) -> T:
        raise NotImplementedError

    def __set__(self, instance: 'ObjectModelABC', value: T):
        raise NotImplementedError

    def __set_name__(self, owner, name):
        raise NotImplementedError

    def __delete__(self, instance):
        raise NotImplementedError

    def serialize(self, instance: 'ObjectModelABC') -> Any:
        raise NotImplementedError

    def deserialize(self, instance: 'ObjectModelABC', value):
        raise NotImplementedError

    def has_default(self) -> bool:
        raise NotImplementedError

    def has_value(self, instance: 'ObjectModelABC'):
        raise NotImplementedError

    def can_provide_value(self, instance: 'ObjectModelABC'):
        raise NotImplementedError

    def validate(self, instance: Optional['ObjectModelABC'], value: T):
        raise NotImplementedError

    def clear(self, instance):
        raise NotImplementedError


class ObjectModelABC:
    __state__: Dict[str, Any]

    def serialize(self) -> Dict[str, Any]:
        raise NotImplementedError

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> 'ObjectModelABC':
        raise NotImplementedError

    def validate(self):
        raise NotImplementedError




from objectmodel.base import FieldABC, ObjectModelABC
from objectmodel.errors import (
    FieldValidationError,
    DuplicateFieldDefinitionError,
    FieldValueRequiredError
)

from objectmodel.model import ObjectModel, ObjectModelMeta
from objectmodel.fields import (
    NOT_PROVIDED,
    Field,
    ProxyField,
    ObjectField,
    ListCollectionField,
    DictCollectionField
)
from ._version import __version__, __version_info__

__all__ = [
    'NOT_PROVIDED',
    'ObjectModel',
    'Field',
    'ListCollectionField',
    'DictCollectionField',
    'ProxyField',
    'FieldValidationError',
    'DuplicateFieldDefinitionError',
    'FieldValueRequiredError'
]
